# credit to https://github.com/aurelio-labs/funkagent
import functools
import inspect
import re
import typing


def type_mapping(dtype):
    """
    Maps a Python type to a JSON Schema type. If the type is a List, it will return the container type and the
    item type. If the type is not recognized, it will default to string. If the type is a Literal, it will return
    the Literal values as an enum. If the type is an Enum, it will return the Enum values as an enum. If the type
    is a List of Enum or Literal, it will return the values of the Enum or Literal as an enum. If the type is a
    List of a type that is not recognized, it will default to string.
    :param dtype:  The Python type to map.
    :return:  A tuple containing the container type and the item type. If the type is not a container, the container
            type will be None.
    """
    type_map = {
        float: "number",
        int: "integer",
        str: "string"
    }

    # Check if it's a List from the typing module
    if hasattr(dtype, '__origin__') and dtype.__origin__ == list:
        # Assuming only one argument for List (like List[int]), else default to string
        item_type = dtype.__args__[0] if dtype.__args__ else str
        return "array", type_map.get(item_type, "string")

    return None, type_map.get(dtype, "string")


def extract_params(doc_str: str):
    """
    Extracts the parameters from a doc string. The doc string must be in the format of the Google style doc string. See
    https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html for more information.
    :param doc_str:  The doc string to extract the parameters from. Must be in the format of the Google style doc
                    string.
    :return:  A dictionary containing the parameters and their descriptions. The keys are the parameter names and the
            values are the descriptions. If there is no description, the value will be an empty string. If there is
            no doc string, the dictionary will be empty.
    """
    # split doc string by newline, skipping empty lines
    params_str = [line for line in doc_str.split("\n") if line.strip()]

    params = {}
    current_param_name = None
    current_param_desc = []

    for line in params_str:
        # If we encounter a return annotation, we end processing for the current parameter
        if line.strip().startswith(':return:'):
            if current_param_name:
                params[current_param_name] = ' '.join(current_param_desc).strip()
                current_param_name = None
                current_param_desc = []
            continue

        # we only look at lines starting with ':param'
        if line.strip().startswith(':param'):
            # Save the previous parameter and its description (if any)
            if current_param_name:
                params[current_param_name] = ' '.join(current_param_desc).strip()
                current_param_desc = []

            param_match = re.findall(r'(?<=:param )\w+', line)
            if param_match:
                current_param_name = param_match[0]
                desc_match = line.replace(f":param {current_param_name}:", "").strip()
                # if there is a description, store it
                if desc_match:
                    current_param_desc.append(desc_match)
        elif current_param_name:  # We're collecting multiline descriptions
            current_param_desc.append(line.strip())

    # Save the last parameter and its description (if any)
    if current_param_name:
        params[current_param_name] = ' '.join(current_param_desc).strip()

    return params


def param_to_title(param_name: str) -> str:
    """Converts a parameter name to title format.
    :param param_name:  The parameter name to convert.
    :return:  The converted parameter name.
    """
    return param_name.replace('_', ' ').title()


def extract_enum_values(dtype) -> typing.Optional[typing.List[str]]:
    """
    Extracts the values of an Enum or Literal type. If the type is not an Enum or Literal, it will return None.
    :param dtype:  The type to extract the values from.
    :return:  A list of the values of the Enum or Literal type. If the type is not an Enum or Literal, it will return
            None. If the type is a List of Enum or Literal, it will return the values of the Enum or Literal as a list.
    """
    # Check if it's an Enum type
    if dtype.__class__.__name__ == 'EnumMeta':
        return [e.value for e in dtype]

        # Check if it's a typing.Literal type (or similar constructs in the typing module)
    elif hasattr(dtype, '__origin__') and dtype.__origin__ == typing.Literal:
        return list(dtype.__args__)

        # Adjusting the check here for Python 3.8:
    elif getattr(dtype, '__origin__', None) == list:
        inner_dtype = dtype.__args__[0] if hasattr(dtype, '__args__') else None

        # For this example, I'm just checking if it's an Enum or a Literal type inside a List.
        if inner_dtype and inner_dtype.__class__.__name__ == 'EnumMeta':
            return [e.value for e in inner_dtype]
        elif hasattr(inner_dtype, '__origin__') and inner_dtype.__origin__ == typing.Literal:
            return list(inner_dtype.__args__)

    return None


def extract_description(func_doc):
    """
    Extracts the description from a doc string. The doc string must be in the format of the Google style doc string. See
    https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html for more information.
    :param func_doc:  The doc string to extract the description from. Must be in the format of the Google style doc
                    string.
    :return:  The description of the doc string. If there is no description, it will return an empty string. If there
            is no doc string, it will return an empty string. If the doc string is not in the Google style format, it
            will return an empty string.
    """
    lines = func_doc.split("\n")
    description_lines = []

    for line in lines:
        if line.strip().startswith(":"):
            break
        description_lines.append(line.strip())

    return ' '.join(description_lines).strip()


def func_metadata(obj):
    """Extracts metadata from a function or a class with a call method. The metadata is in the format of the OpenAPI
        specification. See https://swagger.io/specification/ for more information.
    :param obj:  The function or class to extract the metadata from.
    :return:  A dictionary containing the metadata.
    """

    if isinstance(obj, type):  # If it's a class
        method_to_use = getattr(obj, 'method_name', 'call')
        func = getattr(obj, method_to_use, None)
        if not callable(func):
            raise ValueError(f"Class {obj.__name__} does not have a callable 'call' method.")
        func_name = obj.__name__  # Use the class name
    else:  # If it's a function
        func = obj
        func_name = func.__name__  # Use the function name

    # Handle functools.partial and functools.partialmethod
    fixed_args = {}
    if isinstance(func, (functools.partial, functools.partialmethod)):
        fixed_args = func.keywords or {}
        _func = func.func
        if not fixed_args:
            # noinspection PyUnresolvedReferences
            fixed_args = dict(zip(_func.__code__.co_varnames, func.args))
    else:
        _func = func

    argspec = inspect.getfullargspec(_func)
    func_doc = inspect.getdoc(_func)
    func_description = extract_description(func_doc)
    param_details = extract_params(func_doc) if func_doc else {}

    params = build_params(argspec, fixed_args, param_details)

    _required = determine_required_parameters(argspec, fixed_args)

    return {
        "name": func_name,  # This now correctly uses either the class name or the function name
        "description": func_description,
        "parameters": {
            "type": "object",
            "properties": params,
            "required": _required
        }
    }


def build_params(argspec, fixed_args, param_details):
    """Builds parameters metadata. This is a helper function for func_metadata.
    :param argspec:  The argspec of the function.
    :param fixed_args:  The fixed arguments of the function.
    :param param_details:  The parameter details extracted from the doc string.
    :return:  A dictionary containing the parameters metadata.
    """
    params = {}
    for param_name, param_annotation in argspec.annotations.items():
        if param_name not in fixed_args.keys() and param_name != 'self':
            container_type, mapped_type = type_mapping(param_annotation)

            param_metadata = {
                "title": param_details.get(f"{param_name}_title", param_to_title(param_name)),
                "description": param_details.get(param_name) or "",
                "type": mapped_type
            }

            if container_type == "array":
                param_metadata["type"] = container_type
                param_metadata["items"] = {"type": mapped_type}

            enum_values = extract_enum_values(param_annotation)
            if enum_values:
                param_metadata["enum"] = enum_values

            params[param_name] = param_metadata

    return params


def determine_required_parameters(argspec, fixed_args):
    """Determines the required parameters. This is a helper function for func_metadata.
    :param argspec:  The argspec of the function.
    :param fixed_args:  The fixed arguments of the function.
    :return:  A list of the required parameters.
    """
    _required = [i for i in argspec.args if i not in fixed_args.keys() and i != 'self']

    if argspec.defaults:
        num_defaults = len(argspec.defaults)
        args_with_defaults = argspec.args[-num_defaults:]  # Get the names of arguments with defaults

        _required = [arg for arg in _required if arg not in args_with_defaults]  # Filter out args with defaults

    return _required



import os
import sys

# use this to add the nimbusagent package to the python path
sys.path.append(os.path.join(os.path.dirname(__file__), '..', '..'))

import textwrap
from datetime import datetime
from dotenv import load_dotenv
from nimbusagent.agent.streaming import StreamingAgent
from simple_spinner import SimpleSpinner

MAX_LINE_LENGTH = 100
AGENT_NAME = 'Jack'
MODEL_NAME = 'gpt-4-0613'
TEMPERATURE = 0.5
COLORS = {
    "orange": "\033[38;5;208m",
    "blue": "\033[34m",
    "bold": '\033[1m',
    "reset": '\033[0m'
}
CURSORS = {
    "show": '\033[?25h',
    "hide": '\033[?25l',
    "block": '\033[2 q',
    "underline": '\033[4 q',
    "ibeam": '\033[6 q'
}
INIT_PROMPT = f"Hello I am {AGENT_NAME} a helpful assistant. Let's chat, ask me a question."

load_dotenv()
openai_api_key = os.getenv("OPENAI_API_KEY")
# Initialize Spinner
spinner = SimpleSpinner(f"{COLORS['orange']}Querying {AGENT_NAME}{COLORS['reset']}")


def clear_terminal():
    print('\033c', end='')


def wrap_text(text, width=80):
    return "\n".join(textwrap.fill(line, width) for line in text.splitlines())


def initialize_agent():
    date_now = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S+00:00")
    sys_prompt = (f"You are {AGENT_NAME}, a useful assistant that likes to chat\n"
                  f"with people. Answer users questions and ask follow up questions\n"
                  f"The current date and time is {date_now}.\n\n")
    my_history = [{"role": "assistant", "content": INIT_PROMPT}]
    return StreamingAgent(system_message=sys_prompt, message_history=my_history,
                          openai_api_key=openai_api_key, model_name=MODEL_NAME,
                          temperature=TEMPERATURE)


def interact_with_chatbot(agent, user_input):
    agent_response = ''
    received_response = False
    current_line_length = 0

    for content in agent.ask(user_input):
        if not received_response:
            spinner.stop()
            sys.stdout.write(f"{CURSORS['show']}{COLORS['bold']}{COLORS['blue']}{AGENT_NAME}:{COLORS['reset']} ")
            received_response = True

        if content:
            agent_response += content
            current_line_length += len(content)
            if current_line_length > MAX_LINE_LENGTH:
                sys.stdout.write("\n")
                current_line_length = len(content)
            sys.stdout.write(content)
            sys.stdout.flush()

    sys.stdout.write("\n\n")
    sys.stdout.flush()
    return agent_response


def main():
    agent = initialize_agent()
    clear_terminal()
    print(f"{COLORS['bold']}{COLORS['blue']}Nimbus:{COLORS['reset']} {INIT_PROMPT}\n")

    while True:
        user_input = input(f"{CURSORS['block']}{COLORS['bold']}{COLORS['orange']}You:{COLORS['reset']} ").strip()
        if user_input.lower() == 'exit':
            print("Goodbye!")
            break

        spinner.update("")
        spinner.start()
        interact_with_chatbot(agent, user_input)


if __name__ == '__main__':
    main()



import logging
import os
from typing import Union, List, Iterable, Any

import numpy as np
from openai import OpenAI

FUNCTIONS_EMBEDDING_MODEL = "text-embedding-ada-002"


def is_query_safe(query: str, api_key=None) -> bool:
    """Returns True if the query is considered safe, False otherwise.
    :param query: The query to check.
    :param api_key: The OpenAI API key to use. Uses the OPENAI_API_KEY environment variable if not provided.
    :return: True if the query is considered safe, False otherwise.
    """
    client = OpenAI(api_key=api_key if api_key else os.environ["OPENAI_API_KEY"])

    try:
        response = client.moderations.create(input=query)

        if response and response.results:
            result = response.results[0]

            if result and result.flagged:
                logging.debug(f"Query '{query}' was flagged by OpenAI's moderation API. {result}")
                return False

        return True

    except Exception as e:
        logging.error(f"An error occurred while checking query safety: {e}")

    return False


def get_embedding(text, model=FUNCTIONS_EMBEDDING_MODEL, api_key=None):
    """Returns the embedding of the given text.
    :param text: The text to get the embedding of.
    :param model: The model to use. Defaults to the text-embedding-3-small model.
    :param api_key: The OpenAI API key to use. Uses the OPENAI_API_KEY environment variable if not provided.
    :return: The embedding of the given text.
    """
    try:
        text = text.replace("\n", " ")
        client = OpenAI(api_key=api_key if api_key else os.environ["OPENAI_API_KEY"])
        embedding = client.embeddings.create(
            input=text,
            model=model)
        return embedding.data[0].embedding
    except Exception as e:
        print(f"An error occurred: {e}")
        return None


def cosine_similarity(a, b):
    """ get cosine similarity of two vector of same dimensions
    :param a: The first vector.
    :param b: The second vector.
    :return: The cosine similarity of the two vectors.
    """
    return np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))


def find_similar_embedding_list(query: str,
                                function_embeddings: list,
                                embeddings_model: str = FUNCTIONS_EMBEDDING_MODEL,
                                k_nearest_neighbors: int = 1,
                                min_similarity: float = 0.1):
    """
    Return the k function descriptions most similar to given query.
    :param embeddings_model:
    :param query: The query to check.
    :param function_embeddings: The list of function embeddings to compare to.
    :param k_nearest_neighbors: The number of nearest neighbors to return.
    :param min_similarity: The minimum cosine similarity to consider a function relevant.
    :return: The k function descriptions most similar to given query.
    """
    if not function_embeddings or len(function_embeddings) == 0 or not query:
        return None

    query_embedding = get_embedding(query, model=embeddings_model)
    if not query_embedding:
        return None

    similarities = []
    for function_embedding in function_embeddings:
        similarity = cosine_similarity(query_embedding, function_embedding['embedding'])
        if similarity >= min_similarity:
            similarities.append({'name': function_embedding['name'], 'similarity': similarity})

    # Sort the results by similarity in descending order (most similar first)
    sorted_similarities = sorted(similarities, key=lambda x: x['similarity'], reverse=True)

    # Return the top k nearest neighbors
    return sorted_similarities[:k_nearest_neighbors]


def combine_lists_unique(list1: Iterable[Any], set2: Union[Iterable[Any], set]) -> List[Any]:
    """Combine two lists, removing duplicates.
    :param list1: The first list.
    :param set2: The second list. This can be a set or any iterable.
    :return: The combined list, with duplicates removed.
    """
    if isinstance(list1, list):
        new_list = list1.copy()
    else:
        new_list = list(list1)  # Convert iterable to list

    for item in set2:
        if item not in new_list:
            new_list.append(item)

    return new_list



import os
from typing import Optional, List, Union, Literal, Dict

import openai
from openai import OpenAI

from nimbusagent.functions.handler import FunctionHandler
from nimbusagent.memory.base import AgentMemory
from nimbusagent.utils.helper import is_query_safe, FUNCTIONS_EMBEDDING_MODEL

SYS_MSG = """You are a helpful assistant."""

MODERATION_FAIL_MSG = """I'm sorry, I can't help you with that as it is not appropriate."""

HAVING_TROUBLE_MSG = """I'm sorry, I'm having trouble understanding you."""
DEFAULT_MODEL_NAME = 'gpt-4-turbo'
DEFAULT_TEMP = 0.1
DEFAULT_SECONDARY_MODEL_NAME = 'gpt-3.5-turbo'


class BaseAgent:
    def __init__(
            self,
            openai_api_key: str = None,
            model_name: str = DEFAULT_MODEL_NAME,
            secondary_model_name: str = DEFAULT_SECONDARY_MODEL_NAME,
            temperature: float = DEFAULT_TEMP,
            max_tokens: int = 500,

            functions: Optional[list] = None,
            functions_embeddings: Optional[List[dict]] = None,
            functions_embeddings_model: str = FUNCTIONS_EMBEDDING_MODEL,
            functions_always_use: Optional[List[str]] = None,
            functions_pattern_groups: Optional[List[dict]] = None,
            functions_k_closest: int = 3,
            function_min_similarity: float = 0.5,

            function_max_tokens: int = 2000,
            use_tool_calls: bool = True,

            system_message: str = SYS_MSG,
            message_history: Optional[List[Dict[str, str]]] = None,

            calling_function_start_callback: Optional[callable] = None,
            calling_function_stop_callback: Optional[callable] = None,

            perform_moderation: bool = True,
            moderation_fail_message: str = MODERATION_FAIL_MSG,

            memory_max_entries: int = 20,
            memory_max_tokens: int = 2000,

            internal_thoughts_max_entries: int = 8,
            loops_max: int = 10,

            send_events: bool = False,
            max_event_size: int = 2000,

            on_complete: callable = None,
    ):
        """
        Base Agent Class for Nimbus Agent

        Args:
            openai_api_key: the OpenAI API key to use
            model_name: The name of the model to use (default: 'gpt-4-0613')
            secondary_model_name: The name of the secondary model to use (default: 'gpt-3.5-turbo')
            temperature: The temperature for the response sampling (default: 0.1)
            functions: The list of functions to use (default: None)
            functions_embeddings: The list of function embeddings to use (default: None)
            functions_embeddings_model: The model to use for function embeddings (default: 'text-embedding-ada-002')
            functions_pattern_groups: The list of function pattern groups to use (default: None)
            functions_k_closest: The number of closest embedding functions to use (default: 3)
            function_min_similarity: The minimum similarity to use for embedding functions (default: 0.5)
            functions_always_use: The list of functions to always use (default: None)
            function_max_tokens: The maximum number of tokens to allow for function call. (default: 2500) 0 = unlimited
            use_tool_calls: True if parallel functions should be allowed (default: True). Functions are being
                            deprecated though tool_calls are still a bit beta, so for now this can be set to
                            False to continue using function calls.
            system_message: The message to send to the user when the agent starts
                            (default: "You are a helpful assistant.")
            message_history: The message history to use (default: None)
            calling_function_start_callback: The callback to call when a function is called (default: None)
            calling_function_stop_callback: The callback to call when a function is stopped (default: None)
            perform_moderation: True if moderation should be performed (default: True)
            moderation_fail_message: The message to send to the user when a message is not appropriate
                                    (default: "I'm sorry, I can't help you with that as it is not appropriate.")
            memory_max_entries: The maximum number of entries to store in the memory (default: 20)
            memory_max_tokens: The maximum number of tokens to store in the memory (default: 2000)
            internal_thoughts_max_entries: The maximum number of entries to store in the internal thoughts (default: 3)
            loops_max: The maximum number of loops to allow (default: 5)
            send_events: True if events should be sent (default: False)
            max_event_size: The maximum size of an event (default: 2000)
            on_complete: The callback to call when the agent completes a response.
                The response is passed to the callable. This can be useful with streaming. (default: None)
        """

        self.client = OpenAI(api_key=openai_api_key if openai_api_key is not None else os.getenv("OPENAI_API_KEY"))

        # self.internal_thoughts: A list that captures the agent's intermediate
        # processing and thoughts during a single 'ask' session. It gets cleared
        # at the beginning of every new 'ask' to ensure no residual information
        # affects the new processing.
        self.internal_thoughts = []
        self.internal_thoughts_max_entries = internal_thoughts_max_entries
        self.model_name = model_name
        self.secondary_model_name = secondary_model_name
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.system_message = None
        self.set_system_message(system_message)
        self.last_response = None
        self.perform_moderation = perform_moderation
        self.moderation_fail_message = moderation_fail_message
        self.loops_max = loops_max
        self.send_events = send_events
        self.max_event_size = max_event_size
        self.calling_function_start_callback = calling_function_start_callback
        self.calling_function_stop_callback = calling_function_stop_callback
        self.on_complete = on_complete

        self.chat_history = AgentMemory(max_messages=memory_max_entries, max_tokens=memory_max_tokens)
        if message_history is not None:
            if self._history_needs_moderation(message_history):
                raise ValueError('The message history contains inappropriate content.')
            self.chat_history.set_chat_history(message_history)

        self.function_handler = self._init_function_handler(
            functions=functions,
            functions_embeddings=functions_embeddings,
            functions_embeddings_model=functions_embeddings_model,
            functions_k_closest=functions_k_closest,
            functions_always_use=functions_always_use,
            functions_pattern_groups=functions_pattern_groups,
            function_max_tokens=function_max_tokens,
            function_min_similarity=function_min_similarity)
        self.use_tool_calls = use_tool_calls

    def set_system_message(self, message: str) -> None:
        """Sets the system message.
        :param message: The system message to set
        """
        self.system_message = {"role": "system", "content": message}

    def _init_function_handler(self,
                               functions: Optional[List],
                               functions_embeddings: Optional[List],
                               functions_embeddings_model: str = FUNCTIONS_EMBEDDING_MODEL,
                               functions_k_closest: int = 3,
                               function_min_similarity: float = 0.5,
                               functions_always_use: Optional[List[str]] = None,
                               functions_pattern_groups: Optional[List[dict]] = None,
                               function_max_tokens: int = 0) -> FunctionHandler:
        """Initializes the function handler.
        Returns a FunctionHandler instance.

        :param functions: The list of functions to use
        :param functions_embeddings: The list of function embeddings to use
        :param functions_k_closest: The number of closest functions to use
        :param functions_always_use: The list of functions to always use
        :param functions_pattern_groups: The list of function pattern groups to use
        :return: A FunctionHandler instance
         """

        return FunctionHandler(
            functions=functions,
            embeddings=functions_embeddings,
            embeddings_model=functions_embeddings_model,
            k_nearest=functions_k_closest,
            min_similarity=function_min_similarity,
            always_use=functions_always_use,
            pattern_groups=functions_pattern_groups,
            calling_function_start_callback=self.calling_function_start_callback,
            calling_function_stop_callback=self.calling_function_stop_callback,
            max_tokens=function_max_tokens,
            chat_history=self.chat_history
        )

    # noinspection PyUnresolvedReferences
    def _create_chat_completion(
            self, messages: list, use_functions: bool = True,
            function_call: Union[str, Literal['auto', 'none']] = 'auto',
            stream=False,
            use_secondary_model: bool = False, force_no_functions: bool = False
    ) -> openai.types.chat.ChatCompletion:

        """Creates a chat completion, streaming or not.
        :param messages: The messages to use
        :param use_functions: True if functions should be used
        :param function_call: The function call to use, 'auto' or 'none' or a function name
        :param stream: True if streaming should be used
        :param use_secondary_model: True if the secondary model should be used
        :param force_no_functions: True if functions should be forced to not be used
        :return: An openai chat completion
        """
        model_name = self.secondary_model_name if use_secondary_model else self.model_name

        if use_functions and self.function_handler.functions and not force_no_functions:
            if self.use_tool_calls:
                # noinspection PyTypeChecker
                res = self.client.chat.completions.create(
                    model=model_name,
                    temperature=self.temperature,
                    messages=messages,
                    tools=self.function_handler.functions_to_tools(),
                    tool_choice=function_call,
                    stream=stream)
            else:
                # noinspection PyTypeChecker
                res = self.client.chat.completions.create(
                    model=model_name,
                    temperature=self.temperature,
                    messages=messages,
                    functions=self.function_handler.functions,
                    function_call=function_call,
                    stream=stream)
        else:
            res = self.client.chat.completions.create(
                model=model_name,
                temperature=self.temperature,
                messages=messages,
                stream=stream)
        return res

    def _history_needs_moderation(self, history: List[Dict[str, str]]) -> bool:
        """Handles history moderation.
        Returns True if the history contains inappropriate content, False otherwise.
        :param history: The history to check
        :return: True if the history contains inappropriate content, False otherwise
        """
        if not self.perform_moderation or not history:
            return False

        content_list = [d['content'] for d in history if 'content' in d]

        return self._needs_moderation(" ".join(content_list))

    def _needs_moderation(self, query: str) -> bool:
        """Checks if a query requires moderation.
        Returns True if the query requires moderation, False otherwise.
        :param query: The query to check
        :return: True if the query requires moderation, False otherwise
        """
        return self.perform_moderation and not is_query_safe(query)

    def _clear_internal_thoughts(self) -> None:
        """Clears the internal thoughts of the agent."""
        self.internal_thoughts = []

    def _append_to_chat_history(self, role: str, content: str) -> None:
        """Appends a new message to the chat history.
        :param role: The role of the message
        :param content: The content of the message
        """
        self.chat_history.append({'role': role, 'content': content})

    # noinspection PyUnresolvedReferences
    def get_last_response(self) -> Optional[Union[openai.types.chat.ChatCompletion, str]]:
        """Returns the last response.
        :return: The last response
        """
        return self.last_response

    def get_chat_history(self) -> List[Dict[str, str]]:
        """Returns the chat history.
        :return: The chat history
        """
        return self.chat_history.get_chat_history()

    def get_functions(self) -> Optional[list]:
        """
        Returns the functions.
        :return: The functions
        """
        return self.function_handler.functions

    def clear_chat_history(self) -> None:
        """Clears the chat history."""
        self.chat_history.clear_chat_history()

    def handle_on_complete(self) -> None:
        """Handles the on_complete callback."""
        if self.on_complete and self.last_response:
            self.on_complete(self.last_response)

    def _clear_last_response(self) -> None:
        """Clears the last response."""
        self.last_response = ""



import os
import unittest
from unittest.mock import patch, Mock

from nimbusagent.functions.handler import FunctionHandler

os.environ['OPENAI_API_KEY'] = "test"


class TestFunctionHandler(unittest.TestCase):

    def setUp(self):
        # Initialize FunctionHandler or other required objects here
        self.handler = FunctionHandler()

    def test_initialization(self):
        # Test if FunctionHandler initializes properly
        self.assertIsNotNone(self.handler)

    @patch("nimbusagent.functions.handler.parser.func_metadata")
    def test_parse_functions(self, mock_func_metadata):
        mock_func_metadata.return_value = {"mock": "data"}
        result = self.handler.parse_functions([Mock()])
        self.assertEqual(result, [{"mock": "data"}])

    def test_get_args(self):
        # Test argument extraction
        args = self.handler.get_args("{}")
        self.assertEqual(args, {})

    # Add more tests for other methods and edge cases


if __name__ == '__main__':
    unittest.main()
